package voidious.utils;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.List;
import java.util.Iterator;

import robocode.AdvancedRobot;
import robocode.util.Utils;

/**
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class DiaUtils {
    public static MovSim moveSimulator;
    public static final boolean IGNORE_WALLS = true;
    public static final boolean OBSERVE_WALL_HITS = false;

    public static Point2D.Double project(Point2D.Double sourceLocation, 
        double angle, double length) {
        return new Point2D.Double(sourceLocation.x + Math.sin(angle) * length,
            sourceLocation.y + Math.cos(angle) * length);
    }

    public static double absoluteBearing(Point2D.Double sourceLocation,
           Point2D.Double target) {
            
        return Math.atan2(target.x - sourceLocation.x, 
            target.y - sourceLocation.y);
    }       

    public static int nonZeroSign(double d) {
        if (d < 0) { return -1; }
        return 1;
    }
    
    public static double square(double d) {
        return d * d;
    }
    
    public static double cube(double d) {
        return d * d * d;
    }
    
    public static double limit(double min, double value, double max) {
        return Math.max(min, Math.min(value, max));
    }

    
    public static double botWidthAimAngle(double distance) {
        return Math.abs(18.0/distance);
    }

    public static int bulletTicksFromPower(double distance, double power) {
        return (int)Math.ceil(distance / (20 - (3*power)));
    }

    public static int bulletTicksFromSpeed(double distance, double speed) {
        return (int)Math.ceil(distance / speed);
    }

    public static void setBackAsFront(AdvancedRobot robot, 
        double goAngleRadians) {

        double angle = Utils.normalRelativeAngle(
            goAngleRadians - robot.getHeadingRadians());

        if (Math.abs(angle) > (Math.PI / 2)) {
            if (angle < 0) {
                robot.setTurnRightRadians(Math.PI + angle);
            } else {
                robot.setTurnLeftRadians(Math.PI - angle);
            }

            robot.setBack(100);
        } else {
            if (angle < 0) {
                robot.setTurnLeftRadians(-1 * angle);
           } else {
                robot.setTurnRightRadians(angle);
           }
            robot.setAhead(100);
        }
    }
    
    public static double rollingAverage(double previousValue, double newValue,
        double depth) {
        
        return ((previousValue * depth) + newValue) / (depth + 1);
    }
    
    public static double round(double d, int i) {
        long powerTen = 1;
        
        for (int x = 0; x < i; x++) {
            powerTen *= 10;
        }
        
        return ((double)Math.round(d * powerTen)) / powerTen;
    }
    
    public static Point2D.Double nextLocation(AdvancedRobot robot) {
        if (moveSimulator == null) {
            moveSimulator = new MovSim();
        }

        MovSimStat[] next = moveSimulator.futurePos(1, robot);

        return new Point2D.Double(next[0].x, next[0].y);
    }

    public static Point2D.Double nextLocation(Point2D.Double botLocation,
        double velocity, double heading) {
        
        return new Point2D.Double(
            botLocation.x + (Math.sin(heading)*velocity),
            botLocation.y + (Math.cos(heading)*velocity));
    }

    public static RobotState nextLocation(Point2D.Double botLocation, 
        double velocity, double maxVelocity,
        double headingRadians, double goAngleRadians, long currentTime,
        boolean isSmoothing, boolean ignoreWallHits, double battleFieldWidth,
        double battleFieldHeight) {

        MovSim movSim = DiaUtils.getMovSim();

        double futureTurn = Utils.normalRelativeAngle(
                goAngleRadians - headingRadians);
        double futureDistance;
        
        if (Math.abs(futureTurn) > (Math.PI / 2)) {
            if (futureTurn < 0) { futureTurn = Math.PI + futureTurn; }
            else { futureTurn = -1 * (Math.PI - futureTurn); }
            
            futureDistance = -1000;
        } else {
            futureDistance = 1000;
        }
        
        int extraWallSize = 0;
        
        if (ignoreWallHits) {
            extraWallSize = 50000;
        }   
            
        MovSimStat[] futureMoves = movSim.futurePos(
            1, extraWallSize + botLocation.x, 
            extraWallSize + botLocation.y, velocity, maxVelocity, 
            headingRadians, futureDistance,
            futureTurn, 10.0, extraWallSize * 2 + battleFieldWidth, 
            extraWallSize * 2 + battleFieldHeight);
        
        return new RobotState(
            new Point2D.Double(DiaUtils.round(futureMoves[0].x - extraWallSize, 3),
            DiaUtils.round(futureMoves[0].y - extraWallSize, 3)), futureMoves[0].h,
            futureMoves[0].v, currentTime + 1, isSmoothing);
    }
    
    public static RobotState nextPerpendicularLocation(
            Point2D.Double targetLocation, double absBearingRadians, 
            double enemyVelocity, double enemyHeadingRadians, 
            boolean clockwise, long currentTime, boolean ignoreWallHits) {

        int purelyPerpendicularOffset = 0;
        
        return nextPerpendicularLocation(targetLocation, absBearingRadians, 
            enemyVelocity, enemyHeadingRadians, purelyPerpendicularOffset, 
            clockwise, currentTime, ignoreWallHits);
    }

    public static RobotState nextPerpendicularLocation(
        Point2D.Double targetLocation, double absBearingRadians, 
        double enemyVelocity, double enemyHeadingRadians, double attackAngle,
        boolean clockwise, long currentTime, boolean ignoreWallHits) {

        return nextPerpendicularWallSmoothedLocation(targetLocation,
            absBearingRadians, enemyVelocity, 8.0, enemyHeadingRadians, 
            attackAngle, clockwise, currentTime, null, 0, 0, 0,
            ignoreWallHits);
    }

    public static RobotState nextPerpendicularWallSmoothedLocation(
        Point2D.Double targetLocation, double absBearingRadians, 
        double enemyVelocity, double maxVelocity, double enemyHeadingRadians, 
        double attackAngle, boolean clockwise, long currentTime, 
        Rectangle2D.Double battleField, double bfWidth, double bfHeight, 
        double wallStick, boolean ignoreWallHits) {
            
        int orientation;
        if (clockwise) { orientation = 1; }
        else { orientation = -1; }
        
        double goAngleRadians = Utils.normalRelativeAngle(
            absBearingRadians + (orientation * ((Math.PI / 2) + attackAngle)));

        boolean isSmoothing = false;
        
        if (wallStick != 0 && battleField != null) {
            double smoothedAngle = DiaUtils.wallSmoothing(
                battleField, bfWidth, bfHeight, targetLocation, 
                goAngleRadians, orientation, wallStick);
            
            if (DiaUtils.round(smoothedAngle, 4) 
                != DiaUtils.round(goAngleRadians, 4)) {
                isSmoothing = true;
            }
            
            goAngleRadians = smoothedAngle;
        }
        
        return nextLocation(targetLocation, enemyVelocity, maxVelocity,
            enemyHeadingRadians, goAngleRadians, currentTime,
            isSmoothing, bfWidth, bfHeight, ignoreWallHits);
    }

    public static RobotState nextLocation(
        Point2D.Double targetLocation, double enemyVelocity, double maxVelocity,
        double enemyHeadingRadians, double goAngleRadians, long currentTime,
        boolean isSmoothing, double battleFieldWidth, double battleFieldHeight,
        boolean ignoreWallHits) {

        MovSim movSim = DiaUtils.getMovSim();

        double futureTurn = Utils.normalRelativeAngle(
                goAngleRadians - enemyHeadingRadians);
        double futureDistance;
        
        if (Math.abs(futureTurn) > (Math.PI / 2)) {
            if (futureTurn < 0) { futureTurn = Math.PI + futureTurn; }
            else { futureTurn = -1 * (Math.PI - futureTurn); }
            
            futureDistance = -1000;
        } else {
            futureDistance = 1000;
        }
        
        int extraWallSize = 0;
        
        if (ignoreWallHits) {
            extraWallSize = 50000;
        }   
            
        MovSimStat[] futureMoves = movSim.futurePos(
            1, extraWallSize + targetLocation.x, 
            extraWallSize + targetLocation.y, enemyVelocity, maxVelocity, 
            enemyHeadingRadians, futureDistance,
            futureTurn, 10.0, extraWallSize * 2 + battleFieldWidth, 
            extraWallSize * 2 + battleFieldHeight);
        
        return new RobotState(
            new Point2D.Double(futureMoves[0].x - extraWallSize,
            futureMoves[0].y - extraWallSize), futureMoves[0].h,
            futureMoves[0].v, currentTime + 1, isSmoothing);
    }
        
    public static MovSim getMovSim() {
        if (moveSimulator == null) {
            moveSimulator = new MovSim();
        }

        return moveSimulator;
    }

    public static double orbitalWallDistance(Point2D.Double sourceLocation,
        Point2D.Double targetLocation, double bulletPower, int direction, 
        Rectangle2D.Double fieldRect) {

        return orbitalWallDistance(sourceLocation, targetLocation, bulletPower,
            direction, fieldRect, 1);
    }
    public static double orbitalWallDistance(Point2D.Double sourceLocation,
        Point2D.Double targetLocation, double bulletPower, int direction, 
        Rectangle2D.Double fieldRect, double fudge) {

        double absBearing = 
            DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        double distance = sourceLocation.distance(targetLocation) * fudge;
        double maxAngleRadians = Math.asin(8.0/(20-3.0*bulletPower));

        
        // 1.0 means the max range of orbital movement exactly reaches bounds
        // of battle field
        double wallDistance = 2.0;
        for (int x = 0; x < 200; x++) {
            if (!fieldRect.contains(
                sourceLocation.x + (Math.sin(absBearing +
                    (direction*(x/100.0)*maxAngleRadians))*distance),
                sourceLocation.y + (Math.cos(absBearing +
                    (direction*(x/100.0)*maxAngleRadians))*distance))) {
                wallDistance = x/100.0;
                break;
            }
        }

        return wallDistance;
    }

    public static double directToWallDistance(Point2D.Double targetLocation, 
        double distance, double heading, double bulletPower, 
        Rectangle2D.Double fieldRect) {

        int bulletTicks = DiaUtils.bulletTicksFromPower(distance, bulletPower);
        
        double wallDistance = 2.0;
        for (int x = 0; x < 2 * bulletTicks; x++) {
            if (!fieldRect.contains(
                    targetLocation.x + (Math.sin(heading) * 8.0 * x),
                    targetLocation.y + (Math.cos(heading) * 8.0 * x))) {
                wallDistance = ((double)x)/bulletTicks;
                break;
            }
        }

        return wallDistance;
    }
    
    /**
     * wallSmoothing: do some Voodoo and wall smooth in a very efficient way.
     *   - ...in terms of CPU cycles, not amount of code.
     *   - used to be iterative, which was a lot simpler and more readable,
     *     but far too slow with how much it was called during precise
     *     prediction.
     */
    public static double wallSmoothing(Rectangle2D.Double field, double bfWidth,
        double bfHeight, Point2D.Double startLocation, double startAngleRadians, 
        int orientation, double wallStick) {

/*
        double angle = startAngle;
        _lastWallSmoothAway = false;
        while (!field.contains(x + Math.sin(Math.toRadians(angle))*WALL_STICK,
            y+Math.cos(Math.toRadians(angle))*WALL_STICK)) {
            angle += orientation*smoothNormal*7.0;
            if (smoothNormal == -1) { _lastWallSmoothAway = true; }
        }

        return angle;
*/
        // Trying to do almost exactly the equivalent of the above in more
        // code but less CPU time. The above needs a low increment to work
        // perfectly smoothly, which results in very slow execution.
        //
        // NOTE: The two algorithms can give slightly different results,
        //       but that is mainly because the iterative one never tests a
        //       very specific angle in a corner that would turn up "in bounds";
        //       if it increased the angle var by (1/INFINITY), they'd be the
        //       same (as far as I can tell).

        double angle = startAngleRadians;
        double wallDistanceX = Math.min(startLocation.x - 18, 
            bfWidth - startLocation.x - 18);
        double wallDistanceY = Math.min(startLocation.y - 18, 
            bfHeight - startLocation.y - 18);

        if (wallDistanceX > wallStick && wallDistanceY > wallStick) {
            return startAngleRadians;
        }

        double testX = startLocation.x + (Math.sin(angle)*wallStick);
        double testY = startLocation.y + (Math.cos(angle)*wallStick);
        double testDistanceX = Math.min(testX - 18, bfWidth - testX - 18);
        double testDistanceY = Math.min(testY - 18, bfHeight - testY - 18);

        double adjacent = 0;
        int g = 0;

        while (!field.contains(testX, testY) && g++ < 25) {
            if (angle < 0) { angle += (2*Math.PI); }
            if (testDistanceY < 0 && testDistanceY < testDistanceX) {
                // wall smooth North or South wall
                angle = ((int)((angle + (Math.PI/2)) / Math.PI)) * Math.PI;
                adjacent = Math.abs(wallDistanceY);
            } else if (testDistanceX < 0 && testDistanceX <= testDistanceY) {
                // wall smooth East or West wall
                angle = (((int)(angle / Math.PI)) * Math.PI) + (Math.PI/2);
                adjacent = Math.abs(wallDistanceX);
            }

            angle += orientation *
                (Math.abs(Math.acos(adjacent/wallStick)) + 0.0005);

            testX = startLocation.x + (Math.sin(angle)*wallStick);
            testY = startLocation.y + (Math.cos(angle)*wallStick);
            testDistanceX = Math.min(testX - 18, bfWidth - testX - 18);
            testDistanceY = Math.min(testY - 18, bfHeight - testY - 18);
        }

        return angle;
    }
    
    public static DiaWave findClosestWave(List<DiaWave> waveList,
        Point2D.Double targetLocation, long currentTime, boolean onlySurfable,
        double matchDistanceThreshold) {
                
        double closestDistance = Double.POSITIVE_INFINITY;
        DiaWave closestWave = null;
        Iterator<DiaWave> wavesIterator = waveList.iterator();
        
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            
            double distanceFromTargetToWave =
                (w.sourceLocation.distance(targetLocation) -
                    w.distanceTraveled(currentTime));
            
            if (Math.abs(distanceFromTargetToWave) < matchDistanceThreshold
                && Math.abs(distanceFromTargetToWave) < closestDistance &&
                (!onlySurfable || distanceFromTargetToWave > 0)) {
                
                closestDistance = Math.abs(distanceFromTargetToWave);
                closestWave = w;
            }
        }
        
        return closestWave;     
    }
    
    public static double standardDeviation(double[] values) {
        double avg = average(values);
        
        double sumSquares = 0;
        
        for (int x = 0; x < values.length; x++) {
            sumSquares += square(avg - values[x]);
        }
        
        return Math.sqrt(sumSquares / values.length);
    }
    
    public static double average(double[] values) {
        double sum = 0;
        for (int x = 0; x < values.length; x++) {
            sum += values[x];
        }
        
        return (sum / values.length);
    }
    
    public static double accel(double velocity, double previousVelocity) {
        double accel = velocity - previousVelocity;
        if (previousVelocity == 0.0) {
            accel = Math.abs(accel);
        } else {
            accel *= Math.signum(previousVelocity);
        }
        
        return accel;
    }

}
