package voidious.utils;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import robocode.util.Utils;

/**
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class DiaWave {
    public static final Point2D.Double ORIGIN = new Point2D.Double(0, 0);
    public static final double MAX_GUESSFACTOR = 1.0;
    public static final int CLOCKWISE = 1;
    public static final int COUNTERCLOCKWISE = -1;
    public static final boolean FIRING_WAVE = true;
    public static final boolean NON_FIRING_WAVE = false;
    public static final boolean SURFABLE_WAVE = true;
    public static final boolean ANY_WAVE = false;
    public static final boolean POSITIVE_GUESSFACTOR = true;
    public static final boolean NEGATIVE_GUESSFACTOR = false;
    public static final double NO_CACHED_ESCAPE_ANGLE = -1;
    
    public long fireTime;
    public Point2D.Double sourceLocation;
    public Point2D.Double targetLocation;
    public double absBearing;
    public double bulletPower;
    public double bulletSpeed;
    public int orbitDirection;
    public boolean processedBulletHit;
    public boolean processedPassed;
    public boolean processedWaveBreak;

    public String botName;
    public double targetHeading;
    public double targetRelativeHeading;
    public double targetVelocity;
    public double targetAccel;
    public int targetVelocitySign;
    public double targetDistance;
    public double targetDchangeTime;
    public double targetVchangeTime;
    public double targetWallDistance;
    public double targetRevWallDistance;
    public double targetDl8t;
    public double targetDl20t;
    public double targetDl40t;
    public double gunHeat;
    public int enemiesAlive;
    public Point2D.Double waveBreakLocation;
    public long waveBreakTime;
    public Point2D.Double prevLocation;
    public long prevLocationTime;
    protected Rectangle2D.Double _fieldRect;
    protected double _fieldWidth;
    protected double _fieldHeight;
    
    protected double _cachedPositiveEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    protected double _cachedNegativeEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    public boolean usedNegativeSmoothingMea = false;
    public boolean usedPositiveSmoothingMea = false;

    public DiaWave(Point2D.Double source, Point2D.Double target, 
        Point2D.Double orbitLocation, long time, double power, String name, 
        double heading, double velocity, double accel, int vSign, 
        double distance, double dChange, double vChange, double wallDist1, 
        double wallDist2, double dl8t, double dl20t, double dl40t, int alive, 
        double gh, Rectangle2D.Double field, double width, double height) {
        
        sourceLocation = source;
        targetLocation = target;
        fireTime = time;
        this.setBulletPower(power);
        absBearing = DiaUtils.absoluteBearing(source, target);

        botName = name;
        targetHeading = heading;
        targetVelocity = velocity;
        targetAccel = accel;
        targetVelocitySign = vSign;
        targetDistance = distance;
        targetDchangeTime = dChange;
        targetVchangeTime = vChange;
        targetDl8t = dl8t;
        targetDl20t = dl20t;
        targetDl40t = dl40t;
        enemiesAlive = alive;
        gunHeat = gh;
        
        double orbitRelativeHeading = Utils.normalRelativeAngle(
                effectiveHeading() - 
                    DiaUtils.absoluteBearing(orbitLocation, targetLocation));

        if (orbitRelativeHeading < 0) {
            orbitDirection = COUNTERCLOCKWISE;
        } else {
            orbitDirection = CLOCKWISE;
        }

        targetRelativeHeading = Math.abs(orbitRelativeHeading); 

        this.setWallDistance(orbitLocation, wallDist1, wallDist2);

        _fieldRect = field;
        _fieldWidth = width;
        _fieldHeight = height;

        processedBulletHit = false;
        processedPassed = false;
        processedWaveBreak = false;
        waveBreakLocation = null;
        prevLocation = null;
    }
    
    public void setBulletPower(double power) {
        bulletPower = power;
        bulletSpeed = (20 - (3 * power));
    }

    public void setWallDistance(Point2D.Double orbitLocation, 
        double wallDist1, double wallDist2) {
        
        if (orbitDirection == COUNTERCLOCKWISE && enemiesAlive == 1) {
            // in 1v1, need to switch these if they're orbiting ccw
            targetWallDistance = wallDist2;
            targetRevWallDistance = wallDist1;
        } else {
            // 1v1 and orbiting clockwise 
            // OR melee and ignoring how they're orbiting
            targetWallDistance = wallDist1;
            targetRevWallDistance = wallDist2;
        }
    }
    
    public boolean wavePassed(Point2D.Double lastScanLocation, 
            long lastScanTime, long currentTime) {
        return wavePassed(lastScanLocation, lastScanTime, currentTime, 0);
    }

    // Note: This method really needs to be called every tick for it to
    //       work correctly.
    public boolean wavePassed(Point2D.Double lastScanLocation, 
        long lastScanTime, long currentTime, int offset) {
        
        if (processedPassed) {
            return true;
        } else if (sourceLocation.distanceSq(lastScanLocation) < 
                DiaUtils.square(offset + 
                    (bulletSpeed * (currentTime - fireTime + 1.5))) &&
            lastScanTime == currentTime) {
            if (currentTime - prevLocationTime == 1 ||
                prevLocation == null) {
                waveBreakLocation = lastScanLocation;
                waveBreakTime = currentTime;
            } else {
                double deltaDistance = lastScanLocation.distance(prevLocation) /
                    (currentTime - prevLocationTime);
                double deltaBearing = 
                    DiaUtils.absoluteBearing(prevLocation, lastScanLocation);
                for (long x = prevLocationTime + 1; x <= currentTime; x++) {
                    long interpoTime = x - prevLocationTime;
                    Point2D.Double interpoLocation = 
                        DiaUtils.project(prevLocation, deltaBearing, 
                            (interpoTime) * deltaDistance);
                    if (sourceLocation.distanceSq(interpoLocation) <
                            DiaUtils.square(offset +
                                (bulletSpeed * (x - fireTime + 1.5)))) {
                        waveBreakLocation = interpoLocation;
                        waveBreakTime = x;
                        break;
                    }
                }
                
                if (waveBreakLocation == null) {
                    System.out.println("WARNING: Anomaly in wave break interpolation.");
                    waveBreakLocation = lastScanLocation;
                    waveBreakTime = lastScanTime;
                }
            }
            processedPassed = true;
            return true;
        } else {
            prevLocation = lastScanLocation;
            prevLocationTime = lastScanTime;
            return false;
        }
    }
    
    public boolean wavePassed1v1(Point2D.Double enemyLocation, long currentTime,
        double interceptOffset) {
            
        if (enemyLocation.distanceSq(sourceLocation) < DiaUtils.square(
                bulletSpeed * (currentTime - fireTime) 
                + interceptOffset)) {
            return true;
        } else {
            return false;
        }
    }

    public Point2D.Double waveBreakLocation() {
        if (waveBreakLocation == null) {
            System.out.println("WARNING: Wave.waveBreakLocation() called out of order.");
            return prevLocation;
        } else {
            return waveBreakLocation;
        }
    }
    
    public long waveBreakTime() {
        if (waveBreakLocation == null) {
            return prevLocationTime;
        } else {
            return waveBreakTime;
        }
    }
    
    public long waveBreakBulletTicks() {
        return waveBreakTime() - fireTime;
    }
    
    public double effectiveHeading() {
        return Utils.normalAbsoluteAngle(targetHeading + 
            (targetVelocitySign == 1 ? 0 : Math.PI));
    }
    
    // Note: This should be called after wavePassed is true. If it isn't,
    //       it will just work as though the last seen location is the
    //       wave break location.
    public Point2D.Double displacementVector() {
        double vectorBearing = Utils.normalRelativeAngle(
            DiaUtils.absoluteBearing(targetLocation, waveBreakLocation()) -
                effectiveHeading());
        double vectorDistance = targetLocation.distance(waveBreakLocation()) /
            waveBreakBulletTicks();
        
        return DiaUtils.project(ORIGIN, vectorBearing * orbitDirection, 
            vectorDistance);
    }
    
    public double firingAngleFromDisplacementVector(Point2D.Double dispVector) {
        return Utils.normalAbsoluteAngle(DiaUtils.absoluteBearing(
            sourceLocation, projectLocation(dispVector)));
    }
    
    public double firingAngleFromTargetLocation(Point2D.Double firingTarget) {
        return Utils.normalAbsoluteAngle(DiaUtils.absoluteBearing(
            sourceLocation, firingTarget));
    }
    
    public Point2D.Double projectLocation(Point2D.Double dispVector) {
        double dispAngle = effectiveHeading() +  
            (DiaUtils.absoluteBearing(ORIGIN, dispVector) * orbitDirection);
        double dispDistance = ORIGIN.distance(dispVector);

        Point2D.Double projectedLocation;
        long bulletTicks = DiaUtils.bulletTicksFromSpeed(
            sourceLocation.distance(targetLocation), bulletSpeed);
        long prevBulletTicks = 0;
        long prevPrevBulletTicks;
        int sanityCounter = 0;

        do {
            projectedLocation = DiaUtils.project(targetLocation,
                dispAngle, bulletTicks * dispDistance);
            prevPrevBulletTicks = prevBulletTicks;
            prevBulletTicks = bulletTicks;
            bulletTicks = DiaUtils.bulletTicksFromSpeed(
                sourceLocation.distance(projectedLocation), bulletSpeed);

            if (bulletTicks == prevPrevBulletTicks) {
                projectedLocation = DiaUtils.project(targetLocation, dispAngle, 
                    ((bulletTicks + prevBulletTicks) / 2.0) * dispDistance);
                break;
            }
        } while (bulletTicks != prevBulletTicks && sanityCounter++ < 20);
        
        return projectedLocation; 
    }
    
    public double distanceTraveled(long currentTime) {
        return (currentTime - fireTime) * bulletSpeed;
    }
    
    public double guessFactor(Point2D.Double targetLocation) {
        double newBearingToTarget = 
            DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        double guessAngle = orbitDirection * Utils.normalRelativeAngle( 
            newBearingToTarget - absBearing);
        double maxEscapeAngle = Math.asin(8.0/bulletSpeed);
        
        return guessAngle / maxEscapeAngle;
    }
    
    public double lateralVelocity() {
        return Math.sin(targetRelativeHeading) * Math.abs(targetVelocity);
    }
    
    public double escapeAngleRange() {
        return (preciseEscapeAngle(POSITIVE_GUESSFACTOR) 
                + preciseEscapeAngle(NEGATIVE_GUESSFACTOR));
    }
    
    public double preciseEscapeAngle(boolean positiveGuessFactor) {
        if (positiveGuessFactor) {
            if (_cachedPositiveEscapeAngle == NO_CACHED_ESCAPE_ANGLE) {
                _cachedPositiveEscapeAngle =
                    uncachedPreciseEscapeAngle(positiveGuessFactor) 
                        * MAX_GUESSFACTOR;
            } 
            return _cachedPositiveEscapeAngle;
        } else {
            if (_cachedNegativeEscapeAngle == NO_CACHED_ESCAPE_ANGLE) {
                _cachedNegativeEscapeAngle =
                    uncachedPreciseEscapeAngle(positiveGuessFactor) 
                        * MAX_GUESSFACTOR;
            } 
            return _cachedNegativeEscapeAngle;
        }
    }
    
    public double uncachedPreciseEscapeAngle(boolean positiveGuessFactor) {
        boolean hitWall = false;
        boolean wavePassed = false;
        
        RobotState predictedState = new RobotState( 
            (Point2D.Double)targetLocation.clone(),
            targetHeading, targetVelocity);
        long predictedTime = fireTime;

        boolean clockwisePrediction =
            (orbitDirection == 1 && positiveGuessFactor) ||
            (orbitDirection == -1 && !positiveGuessFactor);
        
        double noSmoothingEscapeAngle = 0;
        double bulletVelocity = this.bulletSpeed;
        
        do {    
            predictedState = DiaUtils.nextPerpendicularLocation(
                predictedState.location, absBearing, 
                predictedState.velocity, predictedState.heading, 
                clockwisePrediction, predictedTime, DiaUtils.IGNORE_WALLS);
            predictedTime = predictedState.time;
            
            if (!_fieldRect.contains(predictedState.location)) {
                hitWall = true;
            } else if (
                wavePassed1v1(predictedState.location, 
                    predictedTime, bulletVelocity)) {
                wavePassed = true;
            }
        } while (!hitWall && !wavePassed);

        noSmoothingEscapeAngle = 
            Math.abs(Utils.normalRelativeAngle(DiaUtils.absoluteBearing(
            sourceLocation, predictedState.location) - absBearing));

        double withSmoothingEscapeAngle = 0;
        
        if (hitWall) {
            double wallSmoothingStick = 80;
            double purelyPerpendicularAttackAngle = 0;
            double fullVelocity = 8.0;
            double orbitAbsBearing = absBearing;
            double bestSmoothingEscapeAngle = 0;
            
            for (int x = 0; x < 3; x++) {
                wavePassed = false;
                predictedState = new RobotState( 
                    (Point2D.Double)targetLocation.clone(),
                    targetHeading, targetVelocity);
                predictedTime = fireTime;
                
                do {    
                    predictedState = 
                        DiaUtils.nextPerpendicularWallSmoothedLocation(
                            predictedState.location, orbitAbsBearing, 
                            predictedState.velocity, fullVelocity, 
                            predictedState.heading, 
                            purelyPerpendicularAttackAngle, clockwisePrediction,
                            predictedTime, _fieldRect, 
                            _fieldWidth, _fieldHeight, 
                            wallSmoothingStick, DiaUtils.OBSERVE_WALL_HITS);
                    predictedTime = predictedState.time;
    //              orbitAbsBearing = DUtils.absoluteBearing(
    //                  sourceLocation, predictedState.location);
                    
                    if (
                        wavePassed1v1(predictedState.location, 
                            predictedTime, bulletVelocity)) {
                        wavePassed = true;
                    }                   
                } while (!wavePassed);
                
                orbitAbsBearing = DiaUtils.absoluteBearing(targetLocation, 
                    predictedState.location) -
                    ((clockwisePrediction ? 1 : -1) * (Math.PI / 2));
                
                bestSmoothingEscapeAngle = Math.max(bestSmoothingEscapeAngle,
                    Math.abs(Utils.normalRelativeAngle(DiaUtils.absoluteBearing(
                    sourceLocation, predictedState.location) - absBearing)));
            }
            withSmoothingEscapeAngle = bestSmoothingEscapeAngle;
        }

        return Math.max(noSmoothingEscapeAngle, withSmoothingEscapeAngle);        
    }
    
    public void clearCachedPreciseEscapeAngles() {
        _cachedPositiveEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
        _cachedNegativeEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    }
}
